#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x5e5a69ea

# Compiled with Coconut version 1.3.1-post_dev17 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import builtins  # line 4
import enum  # line 4
import json  # line 4
import logging  # line 4
import os  # line 4
import shutil  # line 4
sys = _coconut_sys  # line 4
import time  # line 4
import traceback  # line 4
import unittest  # line 4
import uuid  # line 4
from io import BytesIO  # line 5
from io import BufferedRandom  # line 5
from io import TextIOWrapper  # line 5
try:  # Python 3  # line 6
    from unittest import mock  # Python 3  # line 6
except:  # installed via pip  # line 7
    import mock  # installed via pip  # line 7
try:  # only required for mypy  # line 8
    from typing import Any  # only required for mypy  # line 8
    from typing import List  # only required for mypy  # line 8
    from typing import Union  # only required for mypy  # line 8
except:  # line 9
    pass  # line 9

testFolder = os.path.abspath(os.path.join(os.getcwd(), "test"))  # line 11
import configr  # optional dependency  # line 12
os.environ["TEST"] = testFolder  # needed to mock configr library calls in sos  # line 13
import sos  # import of package, not file  # line 14
sos.defaults["defaultbranch"] = "trunk"  # because sos.main() is never called  # line 15

def sync():  # line 17
    if sys.version_info[:2] >= (3, 3):  # line 18
        os.sync()  # line 18


def determineFilesystemTimeResolution() -> 'float':  # line 21
    name = str(uuid.uuid4())  # type: str  # line 22
    with open(name, "w") as fd:  # create temporary file  # line 23
        fd.write("x")  # create temporary file  # line 23
    mt = os.stat(name).st_mtime  # type: float  # get current timestamp  # line 24
    while os.stat(name).st_mtime == mt:  # wait until timestamp modified  # line 25
        time.sleep(0.05)  # to avoid 0.00s bugs (came up some time for unknown reasons)  # line 26
        with open(name, "w") as fd:  # line 27
            fd.write("x")  # line 27
    mt, start, _count = os.stat(name).st_mtime, time.time(), 0  # line 28
    while os.stat(name).st_mtime == mt:  # now cound and measure time until modified again  # line 29
        time.sleep(0.05)  # line 30
        _count += 1  # line 31
        with open(name, "w") as fd:  # line 32
            fd.write("x")  # line 32
    os.unlink(name)  # line 33
    fsprecision = round(time.time() - start, 2)  # type: float  # line 34
    print("File system timestamp precision is %.2fs; wrote to the file %d times during that time" % (fsprecision, _count))  # line 35
    return fsprecision  # line 36


FS_PRECISION = determineFilesystemTimeResolution() * 1.05  # line 39


@_coconut_tco  # line 42
def debugTestRunner(post_mortem=None):  # line 42
    ''' Unittest runner doing post mortem debugging on failing tests. '''  # line 43
    import pdb  # line 44
    if post_mortem is None:  # line 45
        post_mortem = pdb.post_mortem  # line 45
    class DebugTestResult(unittest.TextTestResult):  # line 46
        def addError(_, test, err):  # called before tearDown()  # line 47
            traceback.print_exception(*err)  # line 48
            post_mortem(err[2])  # line 49
            super(DebugTestResult, _).addError(test, err)  # line 50
        def addFailure(_, test, err):  # line 51
            traceback.print_exception(*err)  # line 52
            post_mortem(err[2])  # line 53
            super(DebugTestResult, _).addFailure(test, err)  # line 54
    return _coconut_tail_call(unittest.TextTestRunner, resultclass=DebugTestResult)  # line 55

@_coconut_tco  # line 57
def wrapChannels(func: '_coconut.typing.Callable[..., Any]'):  # line 57
    ''' Wrap function call to capture and return strings emitted on stdout and stderr. '''  # line 58
    oldv = sys.argv  # line 59
    buf = TextIOWrapper(BufferedRandom(BytesIO(b"\0" * 10000)), encoding="utf-8")  # line 60
    sys.stdout = sys.stderr = buf  # line 61
    handler = logging.StreamHandler(buf)  # line 62
    logging.getLogger().addHandler(handler)  # line 63
    try:  # capture output into buf  # line 64
        func()  # capture output into buf  # line 64
    except Exception as E:  # line 65
        buf.write(str(E) + "\n")  # line 65
        traceback.print_exc(file=buf)  # line 65
    except SystemExit as F:  # line 66
        buf.write("EXIT CODE %s" % F.code + "\n")  # line 66
        traceback.print_exc(file=buf)  # line 66
    logging.getLogger().removeHandler(handler)  # line 67
    sys.argv, sys.stdout, sys.stderr = oldv, sys.__stdout__, sys.__stderr__  # TODO when run using pythonw.exe and/or no console, these could be None  # line 68
    buf.seek(0)  # line 69
    return _coconut_tail_call(buf.read)  # line 70

def mockInput(datas: '_coconut.typing.Sequence[str]', func) -> 'Any':  # line 72
    with mock.patch("builtins.input", side_effect=datas):  # line 73
        return func()  # line 73

def setRepoFlag(name: 'str', value: 'bool'):  # line 75
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 76
        flags, branches, config = json.loads(fd.read())  # line 76
    flags[name] = value  # line 77
    with open(sos.metaFolder + os.sep + sos.metaFile, "w") as fd:  # line 78
        fd.write(json.dumps((flags, branches, config)))  # line 78

def checkRepoFlag(name: 'str', flag: '_coconut.typing.Optional[bool]'=None, value: '_coconut.typing.Optional[Any]'=None) -> 'bool':  # line 80
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 81
        flags, branches, config = json.loads(fd.read())  # line 81
    return (name in flags and flags[name] == flag) if flag is not None else (name in config and config[name] == value)  # line 82


class Tests(unittest.TestCase):  # line 85
    ''' Entire test suite. '''  # line 86

    def setUp(_):  # line 88
        for entry in os.listdir(testFolder):  # cannot remove testFolder on Windows when using TortoiseSVN as VCS  # line 89
            resource = os.path.join(testFolder, entry)  # line 90
            shutil.rmtree(resource) if os.path.isdir(resource) else os.unlink(resource)  # line 91
        os.chdir(testFolder)  # line 92


    def assertAllIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]'):  # line 95
        for w in what:  # line 96
            _.assertIn(w, where)  # line 96

    def assertAllNotIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]'):  # line 98
        for w in what:  # line 99
            _.assertNotIn(w, where)  # line 99

    def assertInAll(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 101
        for w in where:  # line 102
            _.assertIn(what, w)  # line 102

    def assertInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 104
        _.assertTrue(any((what in w for w in where)))  # line 104

    def assertNotInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 106
        _.assertFalse(any((what in w for w in where)))  # line 106

    def createFile(_, number: 'Union[int, str]', contents: 'str'="x" * 10, prefix: '_coconut.typing.Optional[str]'=None):  # line 108
        if prefix and not os.path.exists(prefix):  # line 109
            os.makedirs(prefix)  # line 109
        with open(("." if prefix is None else prefix) + os.sep + (("file%d" % number) if isinstance(number, int) else number), "wb") as fd:  # line 110
            fd.write(contents if isinstance(contents, bytes) else contents.encode("cp1252"))  # line 110

    def existsFile(_, number: 'Union[int, str]', expectedContents: 'bytes'=None) -> 'bool':  # line 112
        if not os.path.exists(("." + os.sep + "file%d" % number) if isinstance(number, int) else number):  # line 113
            return False  # line 113
        if expectedContents is None:  # line 114
            return True  # line 114
        with open(("." + os.sep + "file%d" % number) if isinstance(number, int) else number, "rb") as fd:  # line 115
            return fd.read() == expectedContents  # line 115

    def testAccessor(_):  # line 117
        a = sos.Accessor({"a": 1})  # line 118
        _.assertEqual((1, 1), (a["a"], a.a))  # line 119

    def testGetAnyOfmap(_):  # line 121
        _.assertEqual(2, sos.getAnyOfMap({"a": 1, "b": 2}, ["x", "b"]))  # line 122
        _.assertIsNone(sos.getAnyOfMap({"a": 1, "b": 2}, []))  # line 123

    def testAjoin(_):  # line 125
        _.assertEqual("a1a2", sos.ajoin("a", ["1", "2"]))  # line 126
        _.assertEqual("* a\n* b", sos.ajoin("* ", ["a", "b"], "\n"))  # line 127

    def testFindChanges(_):  # line 129
        m = sos.Metadata(os.getcwd())  # line 130
        try:  # line 131
            sos.config(["set", "texttype", "*"])  # line 131
        except SystemExit as E:  # line 132
            _.assertEqual(0, E.code)  # line 132
        try:  # will be stripped from leading paths anyway  # line 133
            sos.config(["set", "ignores", "test/*.cfg;D:\\apps\\*.cfg.bak"])  # will be stripped from leading paths anyway  # line 133
        except SystemExit as E:  # line 134
            _.assertEqual(0, E.code)  # line 134
        m = sos.Metadata(os.getcwd())  # reload from file system  # line 135
        for file in [f for f in os.listdir() if f.endswith(".bak")]:  # remove configuration file  # line 136
            os.unlink(file)  # remove configuration file  # line 136
        _.createFile(1, "1")  # line 137
        m.createBranch(0)  # line 138
        _.assertEqual(1, len(m.paths))  # line 139
        time.sleep(FS_PRECISION)  # time required by filesystem time resolution issues  # line 140
        _.createFile(1, "2")  # modify existing file  # line 141
        _.createFile(2, "2")  # add another file  # line 142
        m.loadCommit(0, 0)  # line 143
        changes = m.findChanges()  # detect time skew  # line 144
        _.assertEqual(1, len(changes.additions))  # line 145
        _.assertEqual(0, len(changes.deletions))  # line 146
        _.assertEqual(1, len(changes.modifications))  # line 147
        m.paths.update(changes.additions)  # line 148
        m.paths.update(changes.modifications)  # line 149
        _.createFile(2, "12")  # modify file again  # line 150
        changes = m.findChanges(0, 1)  # by size, creating new commit  # line 151
        _.assertEqual(0, len(changes.additions))  # line 152
        _.assertEqual(0, len(changes.deletions))  # line 153
        _.assertEqual(1, len(changes.modifications))  # line 154
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1)))  # line 155
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # line 156

    def testPatternPaths(_):  # line 158
        sos.offline(options=["--track"])  # line 159
        os.mkdir("sub")  # line 160
        _.createFile("sub" + os.sep + "file1", "sdfsdf")  # line 161
        sos.add("sub", "sub/file?")  # line 162
        sos.commit("test")  # should pick up sub/file1 pattern  # line 163
        _.assertEqual(2, len(os.listdir(os.path.join(sos.metaFolder, "b0", "r1"))))  # sub/file1 was added  # line 164
        _.createFile(1)  # line 165
        try:  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
            sos.commit("nothing")  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
            _.fail()  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
        except:  # line 167
            pass  # line 167

    def testTokenizeGlobPattern(_):  # line 169
        _.assertEqual([], sos.tokenizeGlobPattern(""))  # line 170
        _.assertEqual([sos.GlobBlock(False, "*", 0)], sos.tokenizeGlobPattern("*"))  # line 171
        _.assertEqual([sos.GlobBlock(False, "*", 0), sos.GlobBlock(False, "???", 1)], sos.tokenizeGlobPattern("*???"))  # line 172
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(True, "x", 2)], sos.tokenizeGlobPattern("x*x"))  # line 173
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(False, "??", 2), sos.GlobBlock(False, "*", 4), sos.GlobBlock(True, "x", 5)], sos.tokenizeGlobPattern("x*??*x"))  # line 174
        _.assertEqual([sos.GlobBlock(False, "?", 0), sos.GlobBlock(True, "abc", 1), sos.GlobBlock(False, "*", 4)], sos.tokenizeGlobPattern("?abc*"))  # line 175

    def testTokenizeGlobPatterns(_):  # line 177
        try:  # because number of literal strings differs  # line 178
            sos.tokenizeGlobPatterns("x*x", "x*")  # because number of literal strings differs  # line 178
            _.fail()  # because number of literal strings differs  # line 178
        except:  # line 179
            pass  # line 179
        try:  # because glob patterns differ  # line 180
            sos.tokenizeGlobPatterns("x*", "x?")  # because glob patterns differ  # line 180
            _.fail()  # because glob patterns differ  # line 180
        except:  # line 181
            pass  # line 181
        try:  # glob patterns differ, regardless of position  # line 182
            sos.tokenizeGlobPatterns("x*", "?x")  # glob patterns differ, regardless of position  # line 182
            _.fail()  # glob patterns differ, regardless of position  # line 182
        except:  # line 183
            pass  # line 183
        sos.tokenizeGlobPatterns("x*", "*x")  # succeeds, because glob patterns match (differ only in position)  # line 184
        sos.tokenizeGlobPatterns("*xb?c", "*x?bc")  # succeeds, because glob patterns match (differ only in position)  # line 185
        try:  # succeeds, because glob patterns match (differ only in position)  # line 186
            sos.tokenizeGlobPatterns("a???b*", "ab???*")  # succeeds, because glob patterns match (differ only in position)  # line 186
            _.fail()  # succeeds, because glob patterns match (differ only in position)  # line 186
        except:  # line 187
            pass  # line 187

    def testConvertGlobFiles(_):  # line 189
        _.assertEqual(["xxayb", "aacb"], [r[1] for r in sos.convertGlobFiles(["axxby", "aabc"], *sos.tokenizeGlobPatterns("a*b?", "*a?b"))])  # line 190
        _.assertEqual(["1qq2ww3", "1abcbx2xbabc3"], [r[1] for r in sos.convertGlobFiles(["qqxbww", "abcbxxbxbabc"], *sos.tokenizeGlobPatterns("*xb*", "1*2*3"))])  # line 191

    def testFolderRemove(_):  # line 193
        m = sos.Metadata(os.getcwd())  # line 194
        _.createFile(1)  # line 195
        _.createFile("a", prefix="sub")  # line 196
        sos.offline()  # line 197
        _.createFile(2)  # line 198
        os.unlink("sub" + os.sep + "a")  # line 199
        os.rmdir("sub")  # line 200
        changes = sos.changes()  # TODO replace by output check  # line 201
        _.assertEqual(1, len(changes.additions))  # line 202
        _.assertEqual(0, len(changes.modifications))  # line 203
        _.assertEqual(1, len(changes.deletions))  # line 204
        _.createFile("a", prefix="sub")  # line 205
        changes = sos.changes()  # line 206
        _.assertEqual(0, len(changes.deletions))  # line 207

    def testSwitchConflict(_):  # line 209
        sos.offline(options=["--strict"])  # (r0)  # line 210
        _.createFile(1)  # line 211
        sos.commit()  # add file (r1)  # line 212
        os.unlink("file1")  # line 213
        sos.commit()  # remove (r2)  # line 214
        _.createFile(1, "something else")  # line 215
        sos.commit()  # (r3)  # line 216
        sos.switch("/1")  # updates file1 - marked as MOD, because mtime was changed  # line 217
        _.existsFile(1, "x" * 10)  # line 218
        sos.switch("/2", ["--force"])  # remove file1 requires --force, because size/content (or mtime in non-strict mode) is different to head of branch  # line 219
        sos.switch("/0")  # do nothing, as file1 is already removed  # line 220
        sos.switch("/1")  # add file1 back  # line 221
        sos.switch("/", ["--force"])  # requires force because changed vs. head of branch  # line 222
        _.existsFile(1, "something else")  # line 223

    def testComputeSequentialPathSet(_):  # line 225
        os.makedirs(sos.branchFolder(0, 0))  # line 226
        os.makedirs(sos.branchFolder(0, 1))  # line 227
        os.makedirs(sos.branchFolder(0, 2))  # line 228
        os.makedirs(sos.branchFolder(0, 3))  # line 229
        os.makedirs(sos.branchFolder(0, 4))  # line 230
        m = sos.Metadata(os.getcwd())  # line 231
        m.branch = 0  # line 232
        m.commit = 2  # line 233
        m.saveBranches()  # line 234
        m.paths = {"./a": sos.PathInfo("", 0, 0, "")}  # line 235
        m.saveCommit(0, 0)  # initial  # line 236
        m.paths["./a"] = sos.PathInfo("", 1, 0, "")  # line 237
        m.saveCommit(0, 1)  # mod  # line 238
        m.paths["./b"] = sos.PathInfo("", 0, 0, "")  # line 239
        m.saveCommit(0, 2)  # add  # line 240
        m.paths["./a"] = sos.PathInfo("", None, 0, "")  # line 241
        m.saveCommit(0, 3)  # del  # line 242
        m.paths["./a"] = sos.PathInfo("", 2, 0, "")  # line 243
        m.saveCommit(0, 4)  # readd  # line 244
        m.commits = {i: sos.CommitInfo(i, 0, None) for i in range(5)}  # line 245
        m.saveBranch(0)  # line 246
        m.computeSequentialPathSet(0, 4)  # line 247
        _.assertEqual(2, len(m.paths))  # line 248

    def testParseRevisionString(_):  # line 250
        m = sos.Metadata(os.getcwd())  # line 251
        m.branch = 1  # line 252
        m.commits = {0: 0, 1: 1, 2: 2}  # line 253
        _.assertEqual((1, 3), m.parseRevisionString("3"))  # line 254
        _.assertEqual((2, 3), m.parseRevisionString("2/3"))  # line 255
        _.assertEqual((1, -1), m.parseRevisionString(None))  # line 256
        _.assertEqual((1, -1), m.parseRevisionString(""))  # line 257
        _.assertEqual((2, -1), m.parseRevisionString("2/"))  # line 258
        _.assertEqual((1, -2), m.parseRevisionString("/-2"))  # line 259
        _.assertEqual((1, -1), m.parseRevisionString("/"))  # line 260

    def testOfflineEmpty(_):  # line 262
        os.mkdir("." + os.sep + sos.metaFolder)  # line 263
        try:  # line 264
            sos.offline("trunk")  # line 264
            _.fail()  # line 264
        except SystemExit as E:  # line 265
            _.assertEqual(1, E.code)  # line 265
        os.rmdir("." + os.sep + sos.metaFolder)  # line 266
        sos.offline("test")  # line 267
        _.assertIn(sos.metaFolder, os.listdir("."))  # line 268
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 269
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 270
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 271
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 272
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 0))))  # only meta data file  # line 273

    def testOfflineWithFiles(_):  # line 275
        _.createFile(1, "x" * 100)  # line 276
        _.createFile(2)  # line 277
        sos.offline("test")  # line 278
        _.assertAllIn(["file1", "file2", sos.metaFolder], os.listdir("."))  # line 279
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 280
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 281
        _.assertAllIn([sos.metaFile, "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2", "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0" + os.sep + "r0"))  # line 282
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 283
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 284
        _.assertEqual(3, len(os.listdir(sos.branchFolder(0, 0))))  # only meta data file plus branch base file copies  # line 285

    def testBranch(_):  # line 287
        _.createFile(1, "x" * 100)  # line 288
        _.createFile(2)  # line 289
        sos.offline("test")  # b0/r0  # line 290
        sos.branch("other")  # b1/r0  # line 291
        _.assertAllIn(["b0", "b1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 292
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b1"))))  # line 293
        _.assertEqual(list(sorted(os.listdir(sos.branchFolder(0, 0)))), list(sorted(os.listdir(sos.branchFolder(1, 0)))))  # line 295
        _.createFile(1, "z")  # modify file  # line 297
        sos.branch()  # b2/r0  branch to unnamed branch with modified file tree contents  # line 298
        _.assertNotEqual(os.stat("." + os.sep + sos.metaFolder + os.sep + "b1" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size, os.stat("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size)  # line 299
        _.createFile(3, "z")  # line 301
        sos.branch("from_last_revision", ["--last", "--stay"])  # b3/r0 create copy of other file1,file2 and don't switch  # line 302
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b3" + os.sep + "r0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0"))))  # line 303
# Check sos.status output which branch is marked


    def testComittingAndChanges(_):  # line 308
        _.createFile(1, "x" * 100)  # line 309
        _.createFile(2)  # line 310
        sos.offline("test")  # line 311
        changes = sos.changes()  # line 312
        _.assertEqual(0, len(changes.additions))  # line 313
        _.assertEqual(0, len(changes.deletions))  # line 314
        _.assertEqual(0, len(changes.modifications))  # line 315
        _.createFile(1, "z")  # size change  # line 316
        changes = sos.changes()  # line 317
        _.assertEqual(0, len(changes.additions))  # line 318
        _.assertEqual(0, len(changes.deletions))  # line 319
        _.assertEqual(1, len(changes.modifications))  # line 320
        sos.commit("message")  # line 321
        _.assertAllIn(["r0", "r1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 322
        _.assertAllIn([sos.metaFile, "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir(sos.branchFolder(0, 1)))  # line 323
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # no further files, only the modified one  # line 324
        _.assertEqual(1, len(sos.changes("/0").modifications))  # vs. explicit revision on current branch  # line 325
        _.assertEqual(1, len(sos.changes("0/0").modifications))  # vs. explicit branch/revision  # line 326
        _.createFile(1, "")  # modify to empty file, mentioned in meta data, but not stored as own file  # line 327
        os.unlink("file2")  # line 328
        changes = sos.changes()  # line 329
        _.assertEqual(0, len(changes.additions))  # line 330
        _.assertEqual(1, len(changes.deletions))  # line 331
        _.assertEqual(1, len(changes.modifications))  # line 332
        sos.commit("modified")  # line 333
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 2))))  # no additional files, only mentions in metadata  # line 334
        try:  # expecting Exit due to no changes  # line 335
            sos.commit("nothing")  # expecting Exit due to no changes  # line 335
            _.fail()  # expecting Exit due to no changes  # line 335
        except:  # line 336
            pass  # line 336

    def testGetBranch(_):  # line 338
        m = sos.Metadata(os.getcwd())  # line 339
        m.branch = 1  # current branch  # line 340
        m.branches = {0: sos.BranchInfo(0, 0, "trunk")}  # line 341
        _.assertEqual(27, m.getBranchByName(27))  # line 342
        _.assertEqual(0, m.getBranchByName("trunk"))  # line 343
        _.assertEqual(1, m.getBranchByName(""))  # split from "/"  # line 344
        _.assertIsNone(m.getBranchByName("unknown"))  # line 345
        m.commits = {0: sos.CommitInfo(0, 0, "bla")}  # line 346
        _.assertEqual(13, m.getRevisionByName("13"))  # line 347
        _.assertEqual(0, m.getRevisionByName("bla"))  # line 348
        _.assertEqual(-1, m.getRevisionByName(""))  # split from "/"  # line 349

    def testTagging(_):  # line 351
        m = sos.Metadata(os.getcwd())  # line 352
        sos.offline()  # line 353
        _.createFile(111)  # line 354
        sos.commit("tag", ["--tag"])  # line 355
        out = wrapChannels(lambda: sos.log()).replace("\r", "").split("\n")  # line 356
        _.assertTrue(any(("|tag" in line and line.endswith("|TAG") for line in out)))  # line 357
        _.createFile(2)  # line 358
        try:  # line 359
            sos.commit("tag")  # line 359
            _.fail()  # line 359
        except:  # line 360
            pass  # line 360
        sos.commit("tag-2", ["--tag"])  # line 361
        out = wrapChannels(lambda: sos.ls(options=["--tags"])).replace("\r", "")  # line 362
        _.assertIn("TAG tag", out)  # line 363

    def testSwitch(_):  # line 365
        _.createFile(1, "x" * 100)  # line 366
        _.createFile(2, "y")  # line 367
        sos.offline("test")  # file1-2  in initial branch commit  # line 368
        sos.branch("second")  # file1-2  switch, having same files  # line 369
        sos.switch("0")  # no change  switch back, no problem  # line 370
        sos.switch("second")  # no change  # switch back, no problem  # line 371
        _.createFile(3, "y")  # generate a file  # line 372
        try:  # uncommited changes detected  # line 373
            sos.switch("test")  # uncommited changes detected  # line 373
            _.fail()  # uncommited changes detected  # line 373
        except SystemExit as E:  # line 374
            _.assertEqual(1, E.code)  # line 374
        sos.commit("Finish")  # file1-3  commit third file into branch second  # line 375
        sos.changes()  # line 376
        sos.switch("test")  # file1-2, remove file3 from file tree  # line 377
        _.assertFalse(_.existsFile(3))  # removed when switching back to test  # line 378
        _.createFile("XXX")  # line 379
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 380
        _.assertIn("File tree has changes", out)  # line 381
        _.assertNotIn("File tree is unchanged", out)  # line 382
        _.assertIn("  * b00   'test'", out)  # line 383
        _.assertIn("    b01 'second'", out)  # line 384
        _.assertIn("(dirty)", out)  # one branch has commits  # line 385
        _.assertIn("(in sync)", out)  # the other doesn't  # line 386
        _.createFile(4, "xy")  # generate a file  # line 387
        sos.switch("second", ["--force"])  # avoids warning on uncommited changes, but keeps file4  # line 388
        _.assertFalse(_.existsFile(4))  # removed when forcedly switching back to test  # line 389
        _.assertTrue(_.existsFile(3))  # was restored from branch's revision r1  # line 390
        os.unlink("." + os.sep + "file1")  # remove old file1  # line 391
        sos.switch("test", ["--force"])  # should restore file1 and remove file3  # line 392
        _.assertTrue(_.existsFile(1))  # was restored from branch's revision r1  # line 393
        _.assertFalse(_.existsFile(3))  # was restored from branch's revision r1  # line 394

    def testAutoDetectVCS(_):  # line 396
        os.mkdir(".git")  # line 397
        sos.offline(sos.vcsBranches[sos.findSosVcsBase()[2]])  # create initial branch  # line 398
        with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 399
            meta = fd.read()  # line 399
        _.assertTrue("\"master\"" in meta)  # line 400
        os.rmdir(".git")  # line 401

    def testUpdate(_):  # line 403
        sos.offline("trunk")  # create initial branch b0/r0  # line 404
        _.createFile(1, "x" * 100)  # line 405
        sos.commit("second")  # create b0/r1  # line 406

        sos.switch("/0")  # go back to b0/r0 - deletes file1  # line 408
        _.assertFalse(_.existsFile(1))  # line 409

        sos.update("/1")  # recreate file1  # line 411
        _.assertTrue(_.existsFile(1))  # line 412

        sos.commit("third", ["--force"])  # force because nothing to commit. should create r2 with same contents as r1, but as differential from r1, not from r0 (= no changes in meta folder)  # line 414
        _.assertTrue(os.path.exists(sos.branchFolder(0, 2)))  # line 415
        _.assertTrue(os.path.exists(sos.branchFolder(0, 2, file=sos.metaFile)))  # line 416
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 2))))  # only meta data file, no differential files  # line 417

        sos.update("/1")  # do nothing, as nothing has changed  # line 419
        _.assertTrue(_.existsFile(1))  # line 420

        _.createFile(2, "y" * 100)  # line 422
#    out = wrapChannels(() -> sos.branch("other"))  # won't comply as there are changes
#    _.assertIn("--force", out)
        sos.branch("other", ["--force"])  # automatically including file 2 (as we are in simple mode)  # line 425
        _.assertTrue(_.existsFile(2))  # line 426
        sos.update("trunk", ["--add"])  # only add stuff  # line 427
        _.assertTrue(_.existsFile(2))  # line 428
        sos.update("trunk")  # nothing to do  # line 429
        _.assertFalse(_.existsFile(2))  # removes file not present in original branch  # line 430

        theirs = b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk"  # line 432
        _.createFile(10, theirs)  # line 433
        mine = b"a\nc\nd\ne\ng\nf\nx\nh\ny\ny\nj"  # missing "b", inserted g, modified g->x, replace x/x -> y/y, removed k  # line 434
        _.createFile(11, mine)  # line 435
        _.assertEqual(b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk", sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.BOTH))  # completely recreated other file  # line 436
        _.assertEqual(b'a\nb\nc\nd\ne\ng\nf\ng\nh\ny\ny\nx\nx\nj\nk', sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.INSERT))  # line 437

    def testUpdate2(_):  # line 439
        _.createFile("test.txt", "x" * 10)  # line 440
        sos.offline("trunk", ["--strict"])  # use strict mode, as timestamp differences are too small for testing  # line 441
        sos.branch("mod")  # line 442
        _.createFile("test.txt", "x" * 5 + "y" * 5)  # line 443
        time.sleep(FS_PRECISION)  # line 444
        sos.commit("mod")  # create b0/r1  # line 445
        sos.switch("trunk", ["--force"])  # should replace contents, force in case some other files were modified (e.g. during working on the code) TODO investigate more  # line 446
        with open("test.txt", "rb") as fd:  # line 447
            _.assertEqual(b"x" * 10, fd.read())  # line 447
        sos.update("mod")  # integrate changes TODO same with ask -> theirs  # line 448
        with open("test.txt", "rb") as fd:  # line 449
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 449
        _.createFile("test.txt", "x" * 10)  # line 450
        mockInput(["t"], lambda: sos.update("mod", ["--ask-lines"]))  # line 451
        with open("test.txt", "rb") as fd:  # line 452
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 452
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 453
        sos.update("mod")  # auto-insert/removes (no intra-line conflict)  # line 454
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 455
        mockInput(["t"], lambda: sos.update("mod", ["--ask"]))  # same as above with interaction -> use theirs (overwrite current file state)  # line 456
        with open("test.txt", "rb") as fd:  # line 457
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 457

    def testIsTextType(_):  # line 459
        m = sos.Metadata(".")  # line 460
        m.c.texttype = ["*.x", "*.md", "*.md.*"]  # line 461
        m.c.bintype = ["*.md.confluence"]  # line 462
        _.assertTrue(m.isTextType("ab.txt"))  # line 463
        _.assertTrue(m.isTextType("./ab.txt"))  # line 464
        _.assertTrue(m.isTextType("bc/ab.txt"))  # line 465
        _.assertFalse(m.isTextType("bc/ab."))  # line 466
        _.assertTrue(m.isTextType("23_3.x.x"))  # line 467
        _.assertTrue(m.isTextType("dfg/dfglkjdf7/test.md"))  # line 468
        _.assertTrue(m.isTextType("./test.md.pdf"))  # line 469
        _.assertFalse(m.isTextType("./test_a.md.confluence"))  # line 470

    def testEolDet(_):  # line 472
        ''' Check correct end-of-line detection. '''  # line 473
        _.assertEqual(b"\n", sos.eoldet(b"a\nb"))  # line 474
        _.assertEqual(b"\r\n", sos.eoldet(b"a\r\nb\r\n"))  # line 475
        _.assertEqual(b"\r", sos.eoldet(b"\ra\rb"))  # line 476
        _.assertAllIn(["Inconsistent", "with "], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\r\na\r\nb\n"))))  # line 477
        _.assertAllIn(["Inconsistent", "without"], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\ra\nnb\n"))))  # line 478
        _.assertIsNone(sos.eoldet(b""))  # line 479
        _.assertIsNone(sos.eoldet(b"sdf"))  # line 480

    def testMerge(_):  # line 482
        ''' Check merge results depending on user options. '''  # line 483
        a = b"a\nb\ncc\nd"  # type: bytes  # line 484
        b = b"a\nb\nee\nd"  # type: bytes  # replaces cc by ee  # line 485
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT))  # one-line block replacement using lineMerge  # line 486
        _.assertEqual(b"a\nb\neecc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.INSERT))  # means insert changes from a into b, but don't replace  # line 487
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.REMOVE))  # means insert changes from a into b, but don't replace  # line 488
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE))  # one-line block replacement using lineMerge  # line 489
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE, charMergeOperation=sos.MergeOperation.REMOVE))  # line 490
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH))  # keeps any changes in b  # line 491
        a = b"a\nb\ncc\nd"  # line 492
        b = b"a\nb\nee\nf\nd"  # replaces cc by block of two lines ee, f  # line 493
        _.assertEqual(b"a\nb\nee\nf\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT))  # multi-line block replacement  # line 494
        _.assertEqual(b"a\nb\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE))  # line 495
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH))  # keeps any changes in b  # line 496
# Test with change + insert
        _.assertEqual(b"a\nb fdcd d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.INSERT))  # line 498
        _.assertEqual(b"a\nb d d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.REMOVE))  # line 499
# Test interactive merge
        a = b"a\nb\nb\ne"  # block-wise replacement  # line 501
        b = b"a\nc\ne"  # line 502
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)))  # line 503
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)))  # line 504
        a = b"a\nb\ne"  # intra-line merge  # line 505
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)))  # line 506
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)))  # line 507

    def testMergeEol(_):  # line 509
        _.assertIn("Differing EOL-styles", wrapChannels(lambda: sos.merge(b"a\nb", b"a\r\nb")))  # expects a warning  # line 510
        _.assertIn(b"a\r\nb", sos.merge(b"a\nb", b"a\r\nb"))  # when in doubt, use "mine" CR-LF  # line 511
        _.assertIn(b"a\nb", sos.merge(b"a\nb", b"a\r\nb", eol=True))  # line 512

    def testPickyMode(_):  # line 514
        ''' Confirm that picky mode reset tracked patterns after commits. '''  # line 515
        sos.offline("trunk", ["--picky"])  # line 516
        changes = sos.changes()  # line 517
        _.assertEqual(0, len(changes.additions))  # do not list any existing file as an addition  # line 518
        sos.add(".", "./file?", ["--force"])  # line 519
        _.createFile(1, "aa")  # line 520
        sos.commit("First")  # add one file  # line 521
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # line 522
        _.createFile(2, "b")  # line 523
        try:  # add nothing, because picky  # line 524
            sos.commit("Second")  # add nothing, because picky  # line 524
        except:  # line 525
            pass  # line 525
        sos.add(".", "./file?")  # line 526
        sos.commit("Third")  # line 527
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # line 528
        out = wrapChannels(lambda: sos.log([])).replace("\r", "")  # line 529
        _.assertIn("  * r2", out)  # line 530
        _.createFile(3, prefix="sub")  # line 531
        sos.add("sub", "sub/file?")  # line 532
        changes = sos.changes()  # line 533
        _.assertEqual(1, len(changes.additions))  # line 534
        _.assertTrue("sub/file3" in changes.additions)  # line 535

    def testTrackedSubfolder(_):  # line 537
        ''' See if patterns for files in sub folders are picked up correctly. '''  # line 538
        os.mkdir("." + os.sep + "sub")  # line 539
        sos.offline("trunk", ["--track"])  # line 540
        _.createFile(1, "x")  # line 541
        _.createFile(1, "x", prefix="sub")  # line 542
        sos.add(".", "./file?")  # add glob pattern to track  # line 543
        sos.commit("First")  # line 544
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 545
        sos.add(".", "sub/file?")  # add glob pattern to track  # line 546
        sos.commit("Second")  # one new file + meta  # line 547
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 548
        os.unlink("file1")  # remove from basefolder  # line 549
        _.createFile(2, "y")  # line 550
        sos.remove(".", "sub/file?")  # line 551
        try:  # raises Exit. TODO test the "suggest a pattern" case  # line 552
            sos.remove(".", "sub/bla")  # raises Exit. TODO test the "suggest a pattern" case  # line 552
            _.fail()  # raises Exit. TODO test the "suggest a pattern" case  # line 552
        except:  # line 553
            pass  # line 553
        sos.commit("Third")  # line 554
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # one new file + meta  # line 555
# TODO also check if /file1 and sub/file1 were removed from index

    def testTrackedMode(_):  # line 558
        ''' Difference in semantics vs simple mode:
          - For remote/other branch we can only know and consider tracked files, thus ignoring all complexity stemming from handling addition of untracked files.
          - For current branch, we can take into account tracked and untracked ones, in theory, but it doesn't make sense.
        In conclusion, using the union of tracking patterns from both sides to find affected files makes sense, but disallow deleting files not present in remote branch.
    '''  # line 563
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 564
        _.createFile(1)  # line 565
        _.createFile("a123a")  # untracked file "a123a"  # line 566
        sos.add(".", "./file?")  # add glob tracking pattern  # line 567
        sos.commit("second")  # versions "file1"  # line 568
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 569
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 570
        _.assertIn("  | ./file?", out)  # line 571

        _.createFile(2)  # untracked file "file2"  # line 573
        sos.commit("third")  # versions "file2"  # line 574
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # one new file + meta file  # line 575

        os.mkdir("." + os.sep + "sub")  # line 577
        _.createFile(3, prefix="sub")  # untracked file "sub/file3"  # line 578
        sos.commit("fourth", ["--force"])  # no tracking pattern matches the subfolder  # line 579
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 3))))  # meta file only, no other tracked path/file  # line 580

        sos.branch("Other")  # second branch containing file1 and file2 tracked by "./file?"  # line 582
        sos.remove(".", "./file?")  # remove tracking pattern, but don't touch previously created and versioned files  # line 583
        sos.add(".", "./a*a")  # add tracking pattern  # line 584
        changes = sos.changes()  # should pick up addition only, because tracked, but not the deletion, as not tracked anymore  # line 585
        _.assertEqual(0, len(changes.modifications))  # line 586
        _.assertEqual(0, len(changes.deletions))  # not tracked anymore, but contained in version history and not removed  # line 587
        _.assertEqual(1, len(changes.additions))  # detected one addition "a123a", but won't recognize untracking files as deletion  # line 588

        sos.commit("Second_2")  # line 590
        _.assertEqual(2, len(os.listdir(sos.branchFolder(1, 1))))  # "a123a" + meta file  # line 591
        _.existsFile(1, b"x" * 10)  # line 592
        _.existsFile(2, b"x" * 10)  # line 593

        sos.switch("test")  # go back to first branch - tracks only "file?", but not "a*a"  # line 595
        _.existsFile(1, b"x" * 10)  # line 596
        _.existsFile("a123a", b"x" * 10)  # line 597

        sos.update("Other")  # integrate tracked files and tracking pattern from second branch into working state of master branch  # line 599
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 600
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 601

        _.createFile("axxxa")  # new file that should be tracked on "test" now that we integrated "Other"  # line 603
        sos.commit("fifth")  # create new revision after integrating updates from second branch  # line 604
        _.assertEqual(3, len(os.listdir(sos.branchFolder(0, 4))))  # one new file from other branch + one new in current folder + meta file  # line 605
        sos.switch("Other")  # switch back to just integrated branch that tracks only "a*a" - shouldn't do anything  # line 606
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 607
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 608
        _.assertFalse(os.path.exists("." + os.sep + "axxxa"))  # because tracked in both branches, but not present in other -> delete in file tree  # line 609
# TODO test switch --meta

    def testLsTracked(_):  # line 612
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 613
        _.createFile(1)  # line 614
        _.createFile("foo")  # line 615
        sos.add(".", "./file*")  # capture one file  # line 616
        sos.ls()  # line 617
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 618
        _.assertInAny("TRK file1  (file*)", out)  # line 619
        _.assertNotInAny("... file1  (file*)", out)  # line 620
        _.assertInAny("... foo", out)  # line 621
        out = sos.safeSplit(wrapChannels(lambda: sos.ls(options=["--patterns"])).replace("\r", ""), "\n")  # line 622
        _.assertInAny("TRK file*", out)  # line 623
        _.createFile("a", prefix="sub")  # line 624
        sos.add("sub", "sub/a")  # line 625
        sos.ls("sub")  # line 626
        _.assertIn("TRK a  (a)", sos.safeSplit(wrapChannels(lambda: sos.ls("sub")).replace("\r", ""), "\n"))  # line 627

    def testLineMerge(_):  # line 629
        _.assertEqual("xabc", sos.lineMerge("xabc", "a bd"))  # line 630
        _.assertEqual("xabxxc", sos.lineMerge("xabxxc", "a bd"))  # line 631
        _.assertEqual("xa bdc", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.INSERT))  # line 632
        _.assertEqual("ab", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.REMOVE))  # line 633

    def testCompression(_):  # TODO test output ratio/advantage, also depending on compress flag set or not  # line 635
        _.createFile(1)  # line 636
        sos.offline("master", options=["--force"])  # line 637
        out = wrapChannels(lambda: sos.changes(options=['--progress'])).replace("\r", "").split("\n")  # line 638
        _.assertFalse(any(("Compression advantage" in line for line in out)))  # simple mode should always print this to stdout  # line 639
        _.assertTrue(_.existsFile(sos.branchFolder(0, 0, file="b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"), b"x" * 10))  # line 640
        setRepoFlag("compress", True)  # was plain = uncompressed before  # line 641
        _.createFile(2)  # line 642
        out = wrapChannels(lambda: sos.commit("Added file2", options=['--progress'])).replace("\r", "").split("\n")  # line 643
        _.assertTrue(any(("Compression advantage" in line for line in out)))  # line 644
        _.assertTrue(_.existsFile(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # exists  # line 645
        _.assertFalse(_.existsFile(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2"), b"x" * 10))  # but is compressed instead  # line 646

    def testLocalConfig(_):  # line 648
        sos.offline("bla", options=[])  # line 649
        try:  # line 650
            sos.config(["set", "ignores", "one;two"], options=["--local"])  # line 650
        except SystemExit as E:  # line 651
            _.assertEqual(0, E.code)  # line 651
        _.assertTrue(checkRepoFlag("ignores", value=["one", "two"]))  # line 652

    def testConfigVariations(_):  # line 654
        def makeRepo():  # line 655
            try:  # line 656
                os.unlink("file1")  # line 656
            except:  # line 657
                pass  # line 657
            sos.offline("master", options=["--force"])  # line 658
            _.createFile(1)  # line 659
            sos.commit("Added file1")  # line 660
        try:  # line 661
            sos.config(["set", "strict", "on"])  # line 661
        except SystemExit as E:  # line 662
            _.assertEqual(0, E.code)  # line 662
        makeRepo()  # line 663
        _.assertTrue(checkRepoFlag("strict", True))  # line 664
        try:  # line 665
            sos.config(["set", "strict", "off"])  # line 665
        except SystemExit as E:  # line 666
            _.assertEqual(0, E.code)  # line 666
        makeRepo()  # line 667
        _.assertTrue(checkRepoFlag("strict", False))  # line 668
        try:  # line 669
            sos.config(["set", "strict", "yes"])  # line 669
        except SystemExit as E:  # line 670
            _.assertEqual(0, E.code)  # line 670
        makeRepo()  # line 671
        _.assertTrue(checkRepoFlag("strict", True))  # line 672
        try:  # line 673
            sos.config(["set", "strict", "no"])  # line 673
        except SystemExit as E:  # line 674
            _.assertEqual(0, E.code)  # line 674
        makeRepo()  # line 675
        _.assertTrue(checkRepoFlag("strict", False))  # line 676
        try:  # line 677
            sos.config(["set", "strict", "1"])  # line 677
        except SystemExit as E:  # line 678
            _.assertEqual(0, E.code)  # line 678
        makeRepo()  # line 679
        _.assertTrue(checkRepoFlag("strict", True))  # line 680
        try:  # line 681
            sos.config(["set", "strict", "0"])  # line 681
        except SystemExit as E:  # line 682
            _.assertEqual(0, E.code)  # line 682
        makeRepo()  # line 683
        _.assertTrue(checkRepoFlag("strict", False))  # line 684
        try:  # line 685
            sos.config(["set", "strict", "true"])  # line 685
        except SystemExit as E:  # line 686
            _.assertEqual(0, E.code)  # line 686
        makeRepo()  # line 687
        _.assertTrue(checkRepoFlag("strict", True))  # line 688
        try:  # line 689
            sos.config(["set", "strict", "false"])  # line 689
        except SystemExit as E:  # line 690
            _.assertEqual(0, E.code)  # line 690
        makeRepo()  # line 691
        _.assertTrue(checkRepoFlag("strict", False))  # line 692
        try:  # line 693
            sos.config(["set", "strict", "enable"])  # line 693
        except SystemExit as E:  # line 694
            _.assertEqual(0, E.code)  # line 694
        makeRepo()  # line 695
        _.assertTrue(checkRepoFlag("strict", True))  # line 696
        try:  # line 697
            sos.config(["set", "strict", "disable"])  # line 697
        except SystemExit as E:  # line 698
            _.assertEqual(0, E.code)  # line 698
        makeRepo()  # line 699
        _.assertTrue(checkRepoFlag("strict", False))  # line 700
        try:  # line 701
            sos.config(["set", "strict", "enabled"])  # line 701
        except SystemExit as E:  # line 702
            _.assertEqual(0, E.code)  # line 702
        makeRepo()  # line 703
        _.assertTrue(checkRepoFlag("strict", True))  # line 704
        try:  # line 705
            sos.config(["set", "strict", "disabled"])  # line 705
        except SystemExit as E:  # line 706
            _.assertEqual(0, E.code)  # line 706
        makeRepo()  # line 707
        _.assertTrue(checkRepoFlag("strict", False))  # line 708
        try:  # line 709
            sos.config(["set", "strict", "nope"])  # line 709
            _.fail()  # line 709
        except SystemExit as E:  # line 710
            _.assertEqual(1, E.code)  # line 710

    def testLsSimple(_):  # line 712
        _.createFile(1)  # line 713
        _.createFile("foo")  # line 714
        _.createFile("ign1")  # line 715
        _.createFile("ign2")  # line 716
        sos.offline("test")  # set up repo in tracking mode (SVN- or gitless-style)  # line 717
        try:  # define an ignore pattern  # line 718
            sos.config(["set", "ignores", "ign1"])  # define an ignore pattern  # line 718
        except SystemExit as E:  # line 719
            _.assertEqual(0, E.code)  # line 719
        try:  # additional ignore pattern  # line 720
            sos.config(["add", "ignores", "ign2"])  # additional ignore pattern  # line 720
        except SystemExit as E:  # line 721
            _.assertEqual(0, E.code)  # line 721
        try:  # define a list of ignore patterns  # line 722
            sos.config(["set", "ignoresWhitelist", "ign1;ign2"])  # define a list of ignore patterns  # line 722
        except SystemExit as E:  # line 723
            _.assertEqual(0, E.code)  # line 723
        out = wrapChannels(lambda: sos.config(["show"])).replace("\r", "")  # line 724
        _.assertIn("             ignores [global]  ['ign1', 'ign2']", out)  # line 725
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 726
        _.assertInAny('... file1', out)  # line 727
        _.assertInAny('... ign1', out)  # line 728
        _.assertInAny('... ign2', out)  # line 729
        try:  # line 730
            sos.config(["rm", "foo", "bar"])  # line 730
            _.fail()  # line 730
        except SystemExit as E:  # line 731
            _.assertEqual(1, E.code)  # line 731
        try:  # line 732
            sos.config(["rm", "ignores", "foo"])  # line 732
            _.fail()  # line 732
        except SystemExit as E:  # line 733
            _.assertEqual(1, E.code)  # line 733
        try:  # line 734
            sos.config(["rm", "ignores", "ign1"])  # line 734
        except SystemExit as E:  # line 735
            _.assertEqual(0, E.code)  # line 735
        try:  # remove ignore pattern  # line 736
            sos.config(["unset", "ignoresWhitelist"])  # remove ignore pattern  # line 736
        except SystemExit as E:  # line 737
            _.assertEqual(0, E.code)  # line 737
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 738
        _.assertInAny('... ign1', out)  # line 739
        _.assertInAny('IGN ign2', out)  # line 740
        _.assertNotInAny('... ign2', out)  # line 741

    def testWhitelist(_):  # line 743
# TODO test same for simple mode
        _.createFile(1)  # line 745
        sos.defaults.ignores[:] = ["file*"]  # replace in-place  # line 746
        sos.offline("xx", ["--track", "--strict"])  # because nothing to commit due to ignore pattern  # line 747
        sos.add(".", "./file*")  # add tracking pattern for "file1"  # line 748
        sos.commit(options=["--force"])  # attempt to commit the file  # line 749
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 1))))  # only meta data, file1 was ignored  # line 750
        try:  # Exit because dirty  # line 751
            sos.online()  # Exit because dirty  # line 751
            _.fail()  # Exit because dirty  # line 751
        except:  # exception expected  # line 752
            pass  # exception expected  # line 752
        _.createFile("x2")  # add another change  # line 753
        sos.add(".", "./x?")  # add tracking pattern for "file1"  # line 754
        try:  # force beyond dirty flag check  # line 755
            sos.online(["--force"])  # force beyond dirty flag check  # line 755
            _.fail()  # force beyond dirty flag check  # line 755
        except:  # line 756
            pass  # line 756
        sos.online(["--force", "--force"])  # force beyond file tree modifications check  # line 757
        _.assertFalse(os.path.exists(sos.metaFolder))  # line 758

        _.createFile(1)  # line 760
        sos.defaults.ignoresWhitelist[:] = ["file*"]  # line 761
        sos.offline("xx", ["--track"])  # line 762
        sos.add(".", "./file*")  # line 763
        sos.commit()  # should NOT ask for force here  # line 764
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # meta data and "file1", file1 was whitelisted  # line 765

    def testRemove(_):  # line 767
        _.createFile(1, "x" * 100)  # line 768
        sos.offline("trunk")  # line 769
        try:  # line 770
            sos.delete("trunk")  # line 770
            _fail()  # line 770
        except:  # line 771
            pass  # line 771
        _.createFile(2, "y" * 10)  # line 772
        sos.branch("added")  # line 773
        sos.delete("trunk")  # line 774
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # meta data file and "b1"  # line 775
        _.assertTrue(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b1"))  # line 776
        _.assertFalse(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 777
        sos.branch("next")  # line 778
        _.createFile(3, "y" * 10)  # make a change  # line 779
        sos.delete("added", "--force")  # should succeed  # line 780

    def testUsage(_):  # line 782
        try:  # TODO expect sys.exit(0)  # line 783
            sos.usage()  # TODO expect sys.exit(0)  # line 783
            _.fail()  # TODO expect sys.exit(0)  # line 783
        except:  # line 784
            pass  # line 784
        try:  # line 785
            sos.usage(short=True)  # line 785
            _.fail()  # line 785
        except:  # line 786
            pass  # line 786

    def testOnly(_):  # line 788
        _.assertEqual((_coconut.frozenset(("./A", "x/B")), _coconut.frozenset(("./C",))), sos.parseOnlyOptions(".", ["abc", "def", "--only", "A", "--x", "--only", "x/B", "--except", "C", "--only"]))  # line 789
        _.assertEqual(_coconut.frozenset(("B",)), sos.conditionalIntersection(_coconut.frozenset(("A", "B", "C")), _coconut.frozenset(("B", "D"))))  # line 790
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(_coconut.frozenset(), _coconut.frozenset(("B", "D"))))  # line 791
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(None, _coconut.frozenset(("B", "D"))))  # line 792
        sos.offline(os.getcwd(), ["--track", "--strict"])  # line 793
        _.createFile(1)  # line 794
        _.createFile(2)  # line 795
        sos.add(".", "./file1")  # line 796
        sos.add(".", "./file2")  # line 797
        sos.commit(onlys=_coconut.frozenset(("./file1",)))  # line 798
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # only meta and file1  # line 799
        sos.commit()  # adds also file2  # line 800
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # only meta and file1  # line 801
        _.createFile(1, "cc")  # modify both files  # line 802
        _.createFile(2, "dd")  # line 803
        try:  # line 804
            sos.config(["set", "texttype", "file2"])  # line 804
        except SystemExit as E:  # line 805
            _.assertEqual(0, E.code)  # line 805
        changes = sos.changes(excps=["./file1"])  # line 806
        _.assertEqual(1, len(changes.modifications))  # only file2  # line 807
        _.assertTrue("./file2" in changes.modifications)  # line 808
        _.assertIn("DIF ./file2", wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 809
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1", "MOD ./file2"], wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 810

    def testDiff(_):  # line 812
        try:  # manually mark this file as "textual"  # line 813
            sos.config(["set", "texttype", "file1"])  # manually mark this file as "textual"  # line 813
        except SystemExit as E:  # line 814
            _.assertEqual(0, E.code)  # line 814
        sos.offline(options=["--strict"])  # line 815
        _.createFile(1)  # line 816
        _.createFile(2)  # line 817
        sos.commit()  # line 818
        _.createFile(1, "sdfsdgfsdf")  # line 819
        _.createFile(2, "12343")  # line 820
        sos.commit()  # line 821
        _.createFile(1, "foobar")  # line 822
        _.createFile(3)  # line 823
        out = wrapChannels(lambda: sos.diff("/-2"))  # compare with r1 (second counting from last which is r2)  # line 824
        _.assertIn("ADD ./file3", out)  # line 825
        _.assertAllIn(["MOD ./file2", "DIF ./file1", "- | 0001 |xxxxxxxxxx|", "+ | 0000 |foobar|"], out)  # line 826
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1"], wrapChannels(lambda: sos.diff("/-2", onlys=_coconut.frozenset(("./file2",)))))  # line 827

    def testReorderRenameActions(_):  # line 829
        result = sos.reorderRenameActions([("123", "312"), ("312", "132"), ("321", "123")], exitOnConflict=False)  # type: Tuple[str, str]  # line 830
        _.assertEqual([("312", "132"), ("123", "312"), ("321", "123")], result)  # line 831
        try:  # line 832
            sos.reorderRenameActions([("123", "312"), ("312", "123")], exitOnConflict=True)  # line 832
            _.fail()  # line 832
        except:  # line 833
            pass  # line 833

    def testMove(_):  # line 835
        sos.offline(options=["--strict", "--track"])  # line 836
        _.createFile(1)  # line 837
        sos.add(".", "./file?")  # line 838
# test source folder missing
        try:  # line 840
            sos.move("sub", "sub/file?", ".", "?file")  # line 840
            _.fail()  # line 840
        except:  # line 841
            pass  # line 841
# test target folder missing: create it
        sos.move(".", "./file?", "sub", "sub/file?")  # line 843
        _.assertTrue(os.path.exists("sub"))  # line 844
        _.assertTrue(os.path.exists("sub/file1"))  # line 845
        _.assertFalse(os.path.exists("file1"))  # line 846
# test move
        sos.move("sub", "sub/file?", ".", "./?file")  # line 848
        _.assertTrue(os.path.exists("1file"))  # line 849
        _.assertFalse(os.path.exists("sub/file1"))  # line 850
# test nothing matches source pattern
        try:  # line 852
            sos.move(".", "a*", ".", "b*")  # line 852
            _.fail()  # line 852
        except:  # line 853
            pass  # line 853
        sos.add(".", "*")  # anything pattern  # line 854
        try:  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 855
            sos.move(".", "a*", ".", "b*")  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 855
            _.fail()  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 855
        except:  # line 856
            pass  # line 856
# test rename no conflict
        _.createFile(1)  # line 858
        _.createFile(2)  # line 859
        _.createFile(3)  # line 860
        sos.add(".", "./file*")  # line 861
        try:  # define an ignore pattern  # line 862
            sos.config(["set", "ignores", "file3;file4"])  # define an ignore pattern  # line 862
        except SystemExit as E:  # line 863
            _.assertEqual(0, E.code)  # line 863
        try:  # line 864
            sos.config(["set", "ignoresWhitelist", "file3"])  # line 864
        except SystemExit as E:  # line 865
            _.assertEqual(0, E.code)  # line 865
        sos.move(".", "./file*", ".", "fi*le")  # line 866
        _.assertTrue(all((os.path.exists("fi%dle" % i) for i in range(1, 4))))  # line 867
        _.assertFalse(os.path.exists("fi4le"))  # line 868
# test rename solvable conflicts
        [_.createFile("%s-%s-%s" % tuple((c for c in n))) for n in ["312", "321", "123", "231"]]  # line 870
#    sos.move("?-?-?")
# test rename unsolvable conflicts
# test --soft option
        sos.remove(".", "./?file")  # was renamed before  # line 874
        sos.add(".", "./?a?b", ["--force"])  # line 875
        sos.move(".", "./?a?b", ".", "./a?b?", ["--force", "--soft"])  # line 876
        _.createFile("1a2b")  # should not be tracked  # line 877
        _.createFile("a1b2")  # should be tracked  # line 878
        sos.commit()  # line 879
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # line 880
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1, file="93b38f90892eb5c57779ca9c0b6fbdf6774daeee3342f56f3e78eb2fe5336c50")))  # a1b2  # line 881
        _.createFile("1a1b1")  # line 882
        _.createFile("1a1b2")  # line 883
        sos.add(".", "?a?b*")  # line 884
        _.assertIn("not unique", wrapChannels(lambda: sos.move(".", "?a?b*", ".", "z?z?")))  # should raise error due to same target name  # line 885
# TODO only rename if actually any files are versioned? or simply what is alife?
# TODO add test if two single question marks will be moved into adjacent characters

    def testHashCollision(_):  # line 889
        sos.offline()  # line 890
        _.createFile(1)  # line 891
        os.mkdir(sos.branchFolder(0, 1))  # line 892
        _.createFile("b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa", prefix=sos.branchFolder(0, 1))  # line 893
        _.createFile(1)  # line 894
        try:  # should exit with error due to collision detection  # line 895
            sos.commit()  # should exit with error due to collision detection  # line 895
            _.fail()  # should exit with error due to collision detection  # line 895
        except SystemExit as E:  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 896
            _.assertEqual(1, E.code)  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 896

    def testFindBase(_):  # line 898
        old = os.getcwd()  # line 899
        try:  # line 900
            os.mkdir("." + os.sep + ".git")  # line 901
            os.makedirs("." + os.sep + "a" + os.sep + sos.metaFolder)  # line 902
            os.makedirs("." + os.sep + "a" + os.sep + "b")  # line 903
            os.chdir("a" + os.sep + "b")  # line 904
            s, vcs, cmd = sos.findSosVcsBase()  # line 905
            _.assertIsNotNone(s)  # line 906
            _.assertIsNotNone(vcs)  # line 907
            _.assertEqual("git", cmd)  # line 908
        finally:  # line 909
            os.chdir(old)  # line 909

# TODO test command line operation --sos vs. --vcs
# check exact output instead of only expected exception/fail


if __name__ == '__main__':  # line 915
    logging.basicConfig(level=logging.DEBUG if '-v' in sys.argv or "true" in [os.getenv("DEBUG", "false").strip().lower(), os.getenv("CI", "false").strip().lower()] else logging.INFO, stream=sys.stderr, format="%(asctime)-23s %(levelname)-8s %(name)s:%(lineno)d | %(message)s" if '-v' in sys.argv or os.getenv("DEBUG", "false") == "true" else "%(message)s")  # line 916
    c = configr.Configr("sos")  # line 917
    c.loadSettings()  # line 918
    if len(c.keys()) > 0:  # line 919
        sos.Exit("Cannot run test suite with existing local SOS user configuration (would affect results)")  # line 919
    unittest.main(testRunner=debugTestRunner() if '-v' in sys.argv and not os.getenv("CI", "false").lower() == "true" else None)  # warnings = "ignore")  # line 920



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xb13e3a7e

# Compiled with Coconut version 1.3.1-post_dev17 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Utiliy functions
import bz2  # line 5
import codecs  # line 5
import difflib  # line 5
import hashlib  # line 5
import logging  # line 5
import os  # line 5
import re  # line 5
sys = _coconut_sys  # line 5
import time  # line 5
START_TIME = time.time()  # line 5
try:  # line 6
    import enum  # line 6
except:  # line 7
    raise Exception("Please install SOS via 'pip install -U sos-vcs[backport]' to get enum support for Python versions prior 3.4")  # line 7
try:  # line 8
    from typing import Any  # only required for mypy  # line 9
    from typing import Callable  # only required for mypy  # line 9
    from typing import Dict  # only required for mypy  # line 9
    from typing import FrozenSet  # only required for mypy  # line 9
    from typing import Generic  # only required for mypy  # line 9
    from typing import IO  # only required for mypy  # line 9
    from typing import List  # only required for mypy  # line 9
    from typing import Sequence  # only required for mypy  # line 9
    from typing import Set  # only required for mypy  # line 9
    from typing import Tuple  # only required for mypy  # line 9
    from typing import Type  # only required for mypy  # line 9
    from typing import TypeVar  # only required for mypy  # line 9
    from typing import Union  # only required for mypy  # line 9
    Number = TypeVar("Number", int, float)  # line 10
    DataType = TypeVar("DataType", BranchInfo, ChangeSet, MergeBlock, PathInfo)  # line 11
except:  # typing not available (prior to Python 3.5)  # line 12
    pass  # typing not available (prior to Python 3.5)  # line 12
try:  # line 13
    import wcwidth  # line 13
except:  # optional dependency  # line 14
    pass  # optional dependency  # line 14


verbose = os.environ.get("DEBUG", "False").lower() == "true" or '--verbose' in sys.argv or '-v' in sys.argv  # imported from utility, and only modified here  # line 17

# Classes
class Accessor(dict):  # line 20
    ''' Dictionary with attribute access. Writing only supported via dictionary access. '''  # line 21
    def __init__(_, mapping: 'Dict[str, Any]'):  # line 22
        dict.__init__(_, mapping)  # line 22
    @_coconut_tco  # line 23
    def __getattribute__(_, name: 'str') -> 'Any':  # line 23
        try:  # line 24
            return _[name]  # line 24
        except:  # line 25
            return _coconut_tail_call(dict.__getattribute__, _, name)  # line 25

class Counter(Generic[Number]):  # line 27
    ''' A simple counter. Can be augmented to return the last value instead. '''  # line 28
    def __init__(_, initial: 'Number'=0):  # line 29
        _.value = initial  # type: Number  # line 29
    def inc(_, by: 'Number'=1) -> 'Number':  # line 30
        _.value += by  # line 30
        return _.value  # line 30

class Logger:  # line 32
    ''' Logger that supports many items. '''  # line 33
    def __init__(_, log):  # line 34
        _._log = log  # line 34
    def debug(_, *s):  # line 35
        _._log.debug(sjoin(*s))  # line 35
    def info(_, *s):  # line 36
        _._log.info(sjoin(*s))  # line 36
    def warn(_, *s):  # line 37
        _._log.warning(sjoin(*s))  # line 37
    def error(_, *s):  # line 38
        _._log.error(sjoin(*s))  # line 38


# Constants
_log = Logger(logging.getLogger(__name__))  # line 42
debug, info, warn, error = _log.debug, _log.info, _log.warn, _log.error  # line 42
UTF8 = "utf_8"  # early used constant, not defined in standard library  # line 43
CONFIGURABLE_FLAGS = ["strict", "track", "picky", "compress"]  # line 44
CONFIGURABLE_LISTS = ["texttype", "bintype", "ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # line 45
GLOBAL_LISTS = ["ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # line 46
TRUTH_VALUES = ["true", "yes", "on", "1", "enable", "enabled"]  # all lower-case normalized  # line 47
FALSE_VALUES = ["false", "no", "off", "0", "disable", "disabled"]  # line 48
PROGRESS_MARKER = ["|", "/", "-", "\\"]  # line 49
metaFolder = ".sos"  # type: str  # line 50
metaFile = ".meta"  # type: str  # line 51
bufSize = 1 << 20  # type: int  # 1 MiB  # line 52
SVN = "svn"  # line 53
SLASH = "/"  # line 54
MEBI = 1 << 20  # line 55
vcsFolders = {".svn": SVN, ".git": "git", ".bzr": "bzr", ".hg": "hg", ".fslckout": "fossil", "_FOSSIL_": "fossil", ".CVS": "cvs"}  # type: Dict[str, str]  # line 56
vcsBranches = {SVN: "trunk", "git": "master", "bzr": "trunk", "hg": "default", "fossil": None, "cvs": None}  # type: Dict[str, _coconut.typing.Optional[str]]  # line 57
lateBinding = Accessor({"verbose": False, "start": 0})  # type: Accessor  # line 58


# Enums
MergeOperation = enum.Enum("MergeOperation", {"INSERT": 1, "REMOVE": 2, "BOTH": 3, "ASK": 4})  # insert remote changes into current, remove remote deletions from current, do both (replicates remote state), or ask per block  # line 62
MergeBlockType = enum.Enum("MergeBlockType", "KEEP INSERT REMOVE REPLACE MOVE")  # modify = intra-line changes, replace = full block replacement  # line 63


# Value types
class BranchInfo(_coconut_NamedTuple("BranchInfo", [("number", 'int'), ("ctime", 'int'), ("name", '_coconut.typing.Optional[str]'), ("inSync", 'bool'), ("tracked", 'List[str]')])):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 67
    __slots__ = ()  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 67
    __ne__ = _coconut.object.__ne__  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 67
    def __new__(_cls, number, ctime, name=None, inSync=False, tracked=[]):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 67
        return _coconut.tuple.__new__(_cls, (number, ctime, name, inSync, tracked))  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 67
# tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place
class CommitInfo(_coconut_NamedTuple("CommitInfo", [("number", 'int'), ("ctime", 'int'), ("message", '_coconut.typing.Optional[str]')])):  # line 68
    __slots__ = ()  # line 68
    __ne__ = _coconut.object.__ne__  # line 68
    def __new__(_cls, number, ctime, message=None):  # line 68
        return _coconut.tuple.__new__(_cls, (number, ctime, message))  # line 68

class PathInfo(_coconut_NamedTuple("PathInfo", [("nameHash", 'str'), ("size", '_coconut.typing.Optional[int]'), ("mtime", 'int'), ("hash", '_coconut.typing.Optional[str]')])):  # size == None means deleted in this revision  # line 69
    __slots__ = ()  # size == None means deleted in this revision  # line 69
    __ne__ = _coconut.object.__ne__  # size == None means deleted in this revision  # line 69
class ChangeSet(_coconut_NamedTuple("ChangeSet", [("additions", 'Dict[str, PathInfo]'), ("deletions", 'Dict[str, PathInfo]'), ("modifications", 'Dict[str, PathInfo]')])):  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 70
    __slots__ = ()  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 70
    __ne__ = _coconut.object.__ne__  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 70
class Range(_coconut_NamedTuple("Range", [("tipe", 'MergeBlockType'), ("indexes", '_coconut.typing.Sequence[int]')])):  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 71
    __slots__ = ()  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 71
    __ne__ = _coconut.object.__ne__  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 71
class MergeBlock(_coconut_NamedTuple("MergeBlock", [("tipe", 'MergeBlockType'), ("lines", 'List[str]'), ("line", 'int'), ("replaces", '_coconut.typing.Optional[MergeBlock]'), ("changes", '_coconut.typing.Optional[Range]')])):  # line 72
    __slots__ = ()  # line 72
    __ne__ = _coconut.object.__ne__  # line 72
    def __new__(_cls, tipe, lines, line, replaces=None, changes=None):  # line 72
        return _coconut.tuple.__new__(_cls, (tipe, lines, line, replaces, changes))  # line 72

class GlobBlock(_coconut_NamedTuple("GlobBlock", [("isLiteral", 'bool'), ("content", 'str'), ("index", 'int')])):  # for file pattern rename/move matching  # line 73
    __slots__ = ()  # for file pattern rename/move matching  # line 73
    __ne__ = _coconut.object.__ne__  # for file pattern rename/move matching  # line 73
class GlobBlock2(_coconut_NamedTuple("GlobBlock2", [("isLiteral", 'bool'), ("content", 'str'), ("matches", 'str')])):  # matching file pattern and input filename for translation  # line 74
    __slots__ = ()  # matching file pattern and input filename for translation  # line 74
    __ne__ = _coconut.object.__ne__  # matching file pattern and input filename for translation  # line 74


# Functions
def printo(s: 'str', nl: 'str'="\n"):  # PEP528 compatibility  # line 78
    sys.stdout.buffer.write((s + nl).encode(sys.stdout.encoding, 'backslashreplace'))  # PEP528 compatibility  # line 78
    sys.stdout.flush()  # PEP528 compatibility  # line 78
def printe(s: 'str', nl: 'str'="\n"):  # line 79
    sys.stderr.buffer.write((s + nl).encode(sys.stderr.encoding, 'backslashreplace'))  # line 79
    sys.stderr.flush()  # line 79
@_coconut_tco  # for py->os access of writing filenames  # PEP 529 compatibility  # line 80
def encode(s: 'str') -> 'bytes':  # for py->os access of writing filenames  # PEP 529 compatibility  # line 80
    return _coconut_tail_call(os.fsencode, s)  # for py->os access of writing filenames  # PEP 529 compatibility  # line 80
@_coconut_tco  # for os->py access of reading filenames  # line 81
def decode(b: 'bytes') -> 'str':  # for os->py access of reading filenames  # line 81
    return _coconut_tail_call(os.fsdecode, b)  # for os->py access of reading filenames  # line 81
try:  # line 82
    import chardet  # https://github.com/chardet/chardet  # line 83
    def detectEncoding(binary: 'bytes') -> 'str':  # line 84
        return chardet.detect(binary)["encoding"]  # line 84
except:  # line 85
    def detectEncoding(binary: 'bytes') -> 'str':  # Guess the encoding  # line 86
        ''' Fallback if chardet library missing. '''  # line 87
        try:  # line 88
            binary.decode(UTF8)  # line 88
            return UTF8  # line 88
        except UnicodeError:  # line 89
            pass  # line 89
        try:  # line 90
            binary.decode("utf_16")  # line 90
            return "utf_16"  # line 90
        except UnicodeError:  # line 91
            pass  # line 91
        try:  # line 92
            binary.decode("cp1252")  # line 92
            return "cp1252"  # line 92
        except UnicodeError:  # line 93
            pass  # line 93
        return "ascii"  # this code will never be reached, as above is an 8-bit charset that always matches  # line 94

def tryOrDefault(func: '_coconut.typing.Callable[[], Any]', default: 'Any') -> 'Any':  # line 96
    try:  # line 97
        return func()  # line 97
    except:  # line 98
        return default  # line 98

@_coconut_tco  # line 100
def wcswidth(string: 'str') -> 'int':  # line 100
    l = 0  # type: int  # line 101
    try:  # line 102
        l = wcwidth.wcswitdh(string)  # line 103
        return len(string) if l < 0 else l  # line 104
    finally:  # line 105
        return _coconut_tail_call(len, string)  # line 105

def removePath(key: 'str', value: 'str') -> 'str':  # line 107
    ''' Cleanup of user-specified global file patterns. '''  # line 108
    return value if value in GLOBAL_LISTS or SLASH not in value else value[value.rindex(SLASH) + 1:]  # line 109

def conditionalIntersection(a: '_coconut.typing.Optional[FrozenSet[str]]', b: 'FrozenSet[str]') -> 'FrozenSet[str]':  # line 111
    return a & b if a else b  # line 111

def dictUpdate(dikt: 'Dict[Any, Any]', by: 'Dict[Any, Any]') -> 'Dict[Any, Any]':  # line 113
    d = {}  # line 113
    d.update(dikt)  # line 113
    d.update(by)  # line 113
    return d  # line 113

def openIt(file: 'str', mode: 'str', compress: 'bool'=False) -> 'IO[bytes]':  # Abstraction for opening both compressed and plain files  # line 115
    return bz2.BZ2File(encode(file), mode) if compress else open(encode(file), mode + "b")  # Abstraction for opening both compressed and plain files  # line 115

def eoldet(file: 'bytes') -> '_coconut.typing.Optional[bytes]':  # line 117
    ''' Determine EOL style from a binary string. '''  # line 118
    lf = file.count(b"\n")  # type: int  # line 119
    cr = file.count(b"\r")  # type: int  # line 120
    crlf = file.count(b"\r\n")  # type: int  # line 121
    if crlf > 0:  # DOS/Windows/Symbian etc.  # line 122
        if lf != crlf or cr != crlf:  # line 123
            warn("Inconsistent CR/NL count with CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 123
        return b"\r\n"  # line 124
    if lf != 0 and cr != 0:  # line 125
        warn("Inconsistent CR/NL count without CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 125
    if lf > cr:  # Linux/Unix  # line 126
        return b"\n"  # Linux/Unix  # line 126
    if cr > lf:  # older 8-bit machines  # line 127
        return b"\r"  # older 8-bit machines  # line 127
    return None  # no new line contained, cannot determine  # line 128

try:  # line 130
    Splittable = TypeVar("Splittable", str, bytes)  # line 130
except:  # line 131
    pass  # line 131
def safeSplit(s: 'Splittable', d: '_coconut.typing.Optional[Splittable]'=None) -> 'List[Splittable]':  # line 132
    return s.split((("\n" if isinstance(s, str) else b"\n") if d is None else d)) if len(s) > 0 else []  # line 132

def ajoin(sep: 'str', seq: '_coconut.typing.Sequence[str]', nl: 'str'="") -> 'str':  # line 134
    return (sep + (nl + sep).join(seq)) if seq else ""  # line 134

@_coconut_tco  # line 136
def sjoin(*s: 'Tuple[Any]') -> 'str':  # line 136
    return _coconut_tail_call(" ".join, [str(e) for e in s if e != ''])  # line 136

@_coconut_tco  # line 138
def hashStr(datas: 'str') -> 'str':  # line 138
    return _coconut_tail_call(hashlib.sha256(datas.encode(UTF8)).hexdigest)  # line 138

def modified(changes: 'ChangeSet', onlyBinary: 'bool'=False) -> 'bool':  # line 140
    return len(changes.additions) > 0 or len(changes.deletions) > 0 or len(changes.modifications) > 0  # line 140

def listindex(lizt: 'Sequence[Any]', what: 'Any', index: 'int'=0) -> 'int':  # line 142
    return lizt[index:].index(what) + index  # line 142

def getTermWidth() -> 'int':  # line 144
    try:  # line 145
        import termwidth  # line 145
    except:  # line 146
        return 80  # line 146
    return termwidth.getTermWidth()[0]  # line 147

def branchFolder(branch: 'int', revision: 'int', base=".", file=None) -> 'str':  # line 149
    return os.path.join(base, metaFolder, "b%d" % branch, "r%d" % revision) + ((os.sep + file) if file else "")  # line 149

def Exit(message: 'str'="", code=1):  # line 151
    printe("[EXIT%s]" % (" %.1fs" % (time.time() - START_TIME) if verbose else "") + (" " + message + "." if message != "" else ""))  # line 151
    sys.exit(code)  # line 151

def exception(E):  # line 153
    ''' Report an exception. '''  # line 154
    printo(str(E))  # line 155
    import traceback  # line 156
    traceback.print_exc()  # line 157
    traceback.print_stack()  # line 158

def hashFile(path: 'str', compress: 'bool', saveTo: '_coconut.typing.Optional[str]'=None) -> 'Tuple[str, int]':  # line 160
    ''' Calculate hash of file contents, and return compressed sized, if in write mode, or zero. '''  # line 161
    _hash = hashlib.sha256()  # line 162
    wsize = 0  # type: int  # line 163
    if saveTo and os.path.exists(encode(saveTo)):  # line 164
        Exit("Hash conflict. Leaving revision in inconsistent state. This should happen only once in a lifetime.")  # line 164
    to = openIt(saveTo, "w", compress) if saveTo else None  # line 165
    with open(encode(path), "rb") as fd:  # line 166
        while True:  # line 167
            buffer = fd.read(bufSize)  # type: bytes  # line 168
            _hash.update(buffer)  # line 169
            if to:  # line 170
                to.write(buffer)  # line 170
            if len(buffer) < bufSize:  # line 171
                break  # line 171
        if to:  # line 172
            to.close()  # line 173
            wsize = os.stat(encode(saveTo)).st_size  # line 174
    return (_hash.hexdigest(), wsize)  # line 175

def getAnyOfMap(map: 'Dict[str, Any]', params: '_coconut.typing.Sequence[str]', default: 'Any'=None) -> 'Any':  # line 177
    ''' Utility. '''  # line 178
    for k, v in map.items():  # line 179
        if k in params:  # line 180
            return v  # line 180
    return default  # line 181

@_coconut_tco  # line 183
def strftime(timestamp: '_coconut.typing.Optional[int]'=None) -> 'str':  # line 183
    return _coconut_tail_call(time.strftime, "%Y-%m-%d %H:%M:%S", time.localtime(timestamp / 1000. if timestamp is not None else None))  # line 183

def detectAndLoad(filename: '_coconut.typing.Optional[str]'=None, content: '_coconut.typing.Optional[bytes]'=None) -> 'Tuple[str, bytes, _coconut.typing.Sequence[str]]':  # line 185
    encoding = None  # type: str  # line 186
    eol = None  # type: bytes  # line 186
    lines = []  # type: _coconut.typing.Sequence[str]  # line 186
    if filename is not None:  # line 187
        with open(encode(filename), "rb") as fd:  # line 188
            content = fd.read()  # line 188
    encoding = (lambda _coconut_none_coalesce_item: sys.getdefaultencoding() if _coconut_none_coalesce_item is None else _coconut_none_coalesce_item)(detectEncoding(content))  # line 189
    eol = eoldet(content)  # line 190
    if filename is not None:  # line 191
        with codecs.open(encode(filename), encoding=encoding) as fd2:  # line 192
            lines = safeSplit(fd2.read(), ((b"\n" if eol is None else eol)).decode(encoding))  # line 192
    elif content is not None:  # line 193
        lines = safeSplit(content.decode(encoding), ((b"\n" if eol is None else eol)).decode(encoding))  # line 194
    else:  # line 195
        return (sys.getdefaultencoding(), b"\n", [])  # line 195
    return (encoding, eol, lines)  # line 196

@_coconut_tco  # line 198
def dataCopy(_tipe: 'Type[DataType]', _old: 'DataType', *_args, **_kwargs) -> 'DataType':  # line 198
    r = _old._asdict()  # line 198
    r.update(**_kwargs)  # line 198
    return _coconut_tail_call(makedata, _tipe, *(list(_args) + [r[field] for field in _old._fields]))  # line 198

def user_block_input(output: 'List[str]'):  # line 200
    sep = input("Enter end-of-text marker (default: <empty line>: ")  # type: str  # line 201
    line = sep  # type: str  # line 201
    while True:  # line 202
        line = input("> ")  # line 203
        if line == sep:  # line 204
            break  # line 204
        output.append(line)  # line 205

@_coconut_tco  # line 207
def merge(file: '_coconut.typing.Optional[bytes]'=None, into: '_coconut.typing.Optional[bytes]'=None, filename: '_coconut.typing.Optional[str]'=None, intoname: '_coconut.typing.Optional[str]'=None, mergeOperation: 'MergeOperation'=MergeOperation.BOTH, charMergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False, eol: 'bool'=False) -> 'Union[bytes, List[MergeBlock]]':  # line 207
    ''' Merges other binary text contents 'file' (or reads file 'filename') into current text contents 'into' (or reads file 'intoname'), returning merged result.
      For update, the other version is assumed to be the "new/added" one, while for diff, the current changes are the ones "added".
      However, change direction markers are insert ("+") for elements only in into, and remove ("-") for elements only in other file (just like the diff marks +/-)
      diffOnly returns detected change blocks only, no text merging
      eol flag will use the other file's EOL marks
      in case of replace block and INSERT strategy, the change will be added **behind** the original
  '''  # line 221
    encoding = None  # type: str  # line 222
    othr = None  # type: _coconut.typing.Sequence[str]  # line 222
    othreol = None  # type: _coconut.typing.Optional[bytes]  # line 222
    curr = None  # type: _coconut.typing.Sequence[str]  # line 222
    curreol = None  # type: _coconut.typing.Optional[bytes]  # line 222
    try:  # load files line-wise and normalize line endings (keep the one of the current file) TODO document  # line 223
        encoding, othreol, othr = detectAndLoad(filename=filename, content=file)  # line 224
        encoding, curreol, curr = detectAndLoad(filename=intoname, content=into)  # line 225
    except Exception as E:  # line 226
        Exit("Cannot merge '%s' into '%s': %r" % (filename, intoname, E))  # line 226
    if None not in [othreol, curreol] and othreol != curreol:  # line 227
        warn("Differing EOL-styles detected during merge. Using current file's style for merged output")  # line 227
    output = list(difflib.Differ().compare(othr, curr))  # type: List[str]  # from generator expression  # line 228
    blocks = []  # type: List[MergeBlock]  # merged result in blocks  # line 229
    tmp = []  # type: List[str]  # block lines  # line 230
    last = " "  # type: str  # line 231
    no = None  # type: int  # line 231
    line = None  # type: str  # line 231
    for no, line in enumerate(output + ["X"]):  # EOF marker (difflib's output will never be "X" alone)  # line 232
        if line[0] == last:  # continue filling current block, no matter what type of block it is  # line 233
            tmp.append(line[2:])  # continue filling current block, no matter what type of block it is  # line 233
            continue  # continue filling current block, no matter what type of block it is  # line 233
        if line == "X" and len(tmp) == 0:  # break if nothing left to do, otherwise perform operation for stored block  # line 234
            break  # break if nothing left to do, otherwise perform operation for stored block  # line 234
        if last == " ":  # block is same in both files  # line 235
            if len(tmp) > 0:  # avoid adding empty keep block  # line 236
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line for line in tmp], line=no - len(tmp)))  # avoid adding empty keep block  # line 236
        elif last == "-":  # may be a pure deletion or part of a replacement (with next block being "+")  # line 237
            blocks.append(MergeBlock(MergeBlockType.REMOVE, [line for line in tmp], line=no - len(tmp)))  # line 238
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.INSERT:  # line 239
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-1].lines, line=no - len(tmp) - 1, replaces=blocks[-2])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 240
                blocks.pop()  # line 241
        elif last == "+":  # may be insertion or replacement (with previous - block)  # line 242
            blocks.append(MergeBlock(MergeBlockType.INSERT, [line for line in tmp], line=no - len(tmp)))  # first, assume simple insertion, then check for replacement  # line 243
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.REMOVE:  #  and len(blocks[-1].lines) == len(blocks[-2].lines):  # requires previous block and same number of lines TODO allow multiple intra-line merge for same-length blocks  # line 244
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, line=no - len(tmp) - 1, replaces=blocks[-1])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 245
                blocks.pop()  # remove TOS due to merging two blocks into replace or modify  # line 246
        elif last == "?":  # marker for intra-line change comment -> add to block info  # line 247
            ilm = getIntraLineMarkers(tmp[0])  # type: Range  # TODO still true? "? " line includes a trailing \n for some reason  # line 248
            blocks[-1] = dataCopy(MergeBlock, blocks[-1], changes=ilm)  # line 249
        last = line[0]  # line 250
        tmp[:] = [line[2:]]  # only keep current line for next block  # line 251
# TODO add code to detect block moves here
    debug("Diff blocks: " + repr(blocks))  # line 253
    if diffOnly:  # line 254
        return blocks  # line 254

# now perform merge operations depending on detected blocks
    output[:] = []  # clean list of strings  # line 257
    for block in blocks:  # line 258
        if block.tipe == MergeBlockType.KEEP:  # line 259
            output.extend(block.lines)  # line 260
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value) or block.tipe == MergeBlockType.REMOVE and (mergeOperation.value & MergeOperation.INSERT.value):  # line 261
            output.extend(block.lines)  # line 263
        elif block.tipe == MergeBlockType.REPLACE:  # complete block replacement  # line 264
            if len(block.lines) == len(block.replaces.lines) == 1:  # one-liner  # line 265
                output.append(lineMerge(block.lines[0], block.replaces.lines[0], mergeOperation=charMergeOperation))  # line 266
            elif mergeOperation == MergeOperation.ASK:  # more than one line: needs user input  # line 267
                printo(ajoin("- ", block.replaces.lines, nl="\n"))  # TODO check +/- in update mode, could be swapped  # line 268
                printo(ajoin("+ ", block.lines, nl="\n"))  # line 269
                while True:  # line 270
                    op = input(" Line replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 271
                    if op in "tb":  # line 272
                        output.extend(block.lines)  # line 272
                        break  # line 272
                    if op in "ib":  # line 273
                        output.extend(block.replaces.lines)  # line 273
                        break  # line 273
                    if op == "m":  # line 274
                        user_block_input(output)  # line 274
                        break  # line 274
            else:  # more than one line and not ask  # line 275
                if mergeOperation == MergeOperation.REMOVE:  # line 276
                    pass  # line 276
                elif mergeOperation == MergeOperation.BOTH:  # line 277
                    output.extend(block.lines)  # line 277
                elif mergeOperation == MergeOperation.INSERT:  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 278
                    output.extend(list(block.replaces.lines) + list(block.lines))  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 278
#  debug("Merge output: " + "; ".join(output))
    nl = othreol if eol else (((b"\n" if othreol is None else othreol) if curreol is None else curreol))  # type: bytes  # line 280
    return _coconut_tail_call(nl.join, [line.encode(encoding) for line in output])  # returning bytes  # line 281
# TODO handle check for more/less lines in found -/+ blocks to find common section and splitting prefix/suffix out

@_coconut_tco  # line 284
def lineMerge(othr: 'str', into: 'str', mergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False) -> 'Union[str, List[MergeBlock]]':  # line 284
    ''' Merges string 'othr' into current string 'into'.
      change direction mark is insert for elements only in into, and remove for elements only in file (according to diff marks +/-)
  '''  # line 287
    out = list(difflib.Differ().compare(othr, into))  # type: List[str]  # line 288
    blocks = []  # type: List[MergeBlock]  # line 289
    for i, line in enumerate(out):  # line 290
        if line[0] == "+":  # line 291
            if i + 1 < len(out) and out[i + 1][0] == "+":  # block will continue  # line 292
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # middle of + block  # line 293
                    blocks[-1].lines.append(line[2])  # add one more character to the accumulating list  # line 294
                else:  # first + in block  # line 295
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 296
            else:  # last line of + block  # line 297
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # end of a block  # line 298
                    blocks[-1].lines.append(line[2])  # line 299
                else:  # single line  # line 300
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 301
                if i >= 1 and blocks[-2].tipe == MergeBlockType.REMOVE:  # previous - and now last in + block creates a replacement block  # line 302
                    blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, i, replaces=blocks[-1])  # line 303
                    blocks.pop()  # line 303
        elif line[0] == "-":  # line 304
            if i > 0 and blocks[-1].tipe == MergeBlockType.REMOVE:  # part of - block  # line 305
                blocks[-1].lines.append(line[2])  # line 306
            else:  # first in block  # line 307
                blocks.append(MergeBlock(MergeBlockType.REMOVE, [line[2]], i))  # line 308
        elif line[0] == " ":  # line 309
            if i > 0 and blocks[-1].tipe == MergeBlockType.KEEP:  # part of block  # line 310
                blocks[-1].lines.append(line[2])  # line 311
            else:  # first in block  # line 312
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line[2]], i))  # line 313
        else:  # line 314
            raise Exception("Cannot parse diff line %r" % line)  # line 314
    blocks[:] = [dataCopy(MergeBlock, block, lines=["".join(block.lines)], replaces=dataCopy(MergeBlock, block.replaces, lines=["".join(block.replaces.lines)]) if block.replaces else None) for block in blocks]  # line 315
    if diffOnly:  # line 316
        return blocks  # line 316
    out[:] = []  # line 317
    for i, block in enumerate(blocks):  # line 318
        if block.tipe == MergeBlockType.KEEP:  # line 319
            out.extend(block.lines)  # line 319
        elif block.tipe == MergeBlockType.REPLACE:  # line 320
            if mergeOperation == MergeOperation.ASK:  # line 321
                printo(ajoin("- ", othr))  # line 322
                printo("- " + (" " * i) + block.replaces.lines[0])  # line 323
                printo("+ " + (" " * i) + block.lines[0])  # line 324
                printo(ajoin("+ ", into))  # line 325
                while True:  # line 326
                    op = input(" Character replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 327
                    if op in "tb":  # line 328
                        out.extend(block.lines)  # line 328
                        break  # line 328
                    if op in "ib":  # line 329
                        out.extend(block.replaces.lines)  # line 329
                        break  # line 329
                    if op == "m":  # line 330
                        user_block_input(out)  # line 330
                        break  # line 330
            else:  # non-interactive  # line 331
                if mergeOperation == MergeOperation.REMOVE:  # line 332
                    pass  # line 332
                elif mergeOperation == MergeOperation.BOTH:  # line 333
                    out.extend(block.lines)  # line 333
                elif mergeOperation == MergeOperation.INSERT:  # line 334
                    out.extend(list(block.replaces.lines) + list(block.lines))  # line 334
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value):  # line 335
            out.extend(block.lines)  # line 335
        elif block.tipe == MergeBlockType.REMOVE and mergeOperation.value & MergeOperation.INSERT.value:  # line 336
            out.extend(block.lines)  # line 336
# TODO ask for insert or remove as well
    return _coconut_tail_call("".join, out)  # line 338

@_coconut_tco  # line 340
def getIntraLineMarkers(line: 'str') -> 'Range':  # line 340
    ''' Return (type, [affected indices]) of "? "-line diff markers ("? " prefix must be removed). difflib never returns mixed markers per line. '''  # line 341
    if "^" in line:  # TODO wrong, needs removal anyway  # line 342
        return _coconut_tail_call(Range, MergeBlockType.REPLACE, [i for i, c in enumerate(line) if c == "^"])  # TODO wrong, needs removal anyway  # line 342
    if "+" in line:  # line 343
        return _coconut_tail_call(Range, MergeBlockType.INSERT, [i for i, c in enumerate(line) if c == "+"])  # line 343
    if "-" in line:  # line 344
        return _coconut_tail_call(Range, MergeBlockType.REMOVE, [i for i, c in enumerate(line) if c == "-"])  # line 344
    return _coconut_tail_call(Range, MergeBlockType.KEEP, [])  # line 345

def findSosVcsBase() -> 'Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str], _coconut.typing.Optional[str]]':  # line 347
    ''' Attempts to find sos and legacy VCS base folders. '''  # line 348
    debug("Detecting root folders...")  # line 349
    path = os.getcwd()  # type: str  # start in current folder, check parent until found or stopped  # line 350
    vcs = (None, None)  # type: Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str]]  # line 351
    while not os.path.exists(encode(os.path.join(path, metaFolder))):  # line 352
        contents = set(os.listdir(path))  # type: Set[str]  # line 353
        vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # type: _coconut.typing.Sequence[str]  # determine VCS type from dot folder  # line 354
        choice = None  # type: _coconut.typing.Optional[str]  # line 355
        if len(vcss) > 1:  # line 356
            choice = SVN if SVN in vcss else vcss[0]  # line 357
            warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 358
        elif len(vcss) > 0:  # line 359
            choice = vcss[0]  # line 359
        if not vcs[0] and choice:  # memorize current repo root  # line 360
            vcs = (path, choice)  # memorize current repo root  # line 360
        new = os.path.dirname(path)  # get parent path  # line 361
        if new == path:  # avoid infinite loop  # line 362
            break  # avoid infinite loop  # line 362
        path = new  # line 363
    if os.path.exists(encode(os.path.join(path, metaFolder))):  # found something  # line 364
        if vcs[0]:  # already detected vcs base and command  # line 365
            return (path, vcs[0], vcs[1])  # already detected vcs base and command  # line 365
        sos = path  # line 366
        while True:  # continue search for VCS base  # line 367
            new = os.path.dirname(path)  # get parent path  # line 368
            if new == path:  # no VCS folder found  # line 369
                return (sos, None, None)  # no VCS folder found  # line 369
            path = new  # line 370
            contents = set(os.listdir(path))  # line 371
            vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # determine VCS type  # line 372
            choice = None  # line 373
            if len(vcss) > 1:  # line 374
                choice = SVN if SVN in vcss else vcss[0]  # line 375
                warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 376
            elif len(vcss) > 0:  # line 377
                choice = vcss[0]  # line 377
            if choice:  # line 378
                return (sos, path, choice)  # line 378
    return (None, vcs[0], vcs[1])  # line 379

def tokenizeGlobPattern(pattern: 'str') -> 'List[GlobBlock]':  # line 381
    index = 0  # type: int  # line 382
    out = []  # type: List[GlobBlock]  # literal = True, first index  # line 383
    while index < len(pattern):  # line 384
        if pattern[index:index + 3] in ("[?]", "[*]", "[[]", "[]]"):  # line 385
            out.append(GlobBlock(False, pattern[index:index + 3], index))  # line 385
            continue  # line 385
        if pattern[index] in "*?":  # line 386
            count = 1  # type: int  # line 387
            while index + count < len(pattern) and pattern[index] == "?" and pattern[index + count] == "?":  # line 388
                count += 1  # line 388
            out.append(GlobBlock(False, pattern[index:index + count], index))  # line 389
            index += count  # line 389
            continue  # line 389
        if pattern[index:index + 2] == "[!":  # line 390
            out.append(GlobBlock(False, pattern[index:pattern.index("]", index + 2) + 1], index))  # line 390
            index += len(out[-1][1])  # line 390
            continue  # line 390
        count = 1  # line 391
        while index + count < len(pattern) and pattern[index + count] not in "*?[":  # line 392
            count += 1  # line 392
        out.append(GlobBlock(True, pattern[index:index + count], index))  # line 393
        index += count  # line 393
    return out  # line 394

def tokenizeGlobPatterns(oldPattern: 'str', newPattern: 'str') -> 'Tuple[_coconut.typing.Sequence[GlobBlock], _coconut.typing.Sequence[GlobBlock]]':  # line 396
    ot = tokenizeGlobPattern(oldPattern)  # type: List[GlobBlock]  # line 397
    nt = tokenizeGlobPattern(newPattern)  # type: List[GlobBlock]  # line 398
#  if len(ot) != len(nt): Exit("Source and target patterns can't be translated due to differing number of parsed glob markers and literal strings")
    if len([o for o in ot if not o.isLiteral]) < len([n for n in nt if not n.isLiteral]):  # line 400
        Exit("Source and target file patterns contain differing number of glob markers and can't be translated")  # line 400
    if any((O.content != N.content for O, N in zip([o for o in ot if not o.isLiteral], [n for n in nt if not n.isLiteral]))):  # line 401
        Exit("Source and target file patterns differ in semantics")  # line 401
    return (ot, nt)  # line 402

def convertGlobFiles(filenames: '_coconut.typing.Sequence[str]', oldPattern: '_coconut.typing.Sequence[GlobBlock]', newPattern: '_coconut.typing.Sequence[GlobBlock]') -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 404
    ''' Converts given filename according to specified file patterns. No support for adjacent glob markers currently. '''  # line 405
    pairs = []  # type: List[Tuple[str, str]]  # line 406
    for filename in filenames:  # line 407
        literals = [l for l in oldPattern if l.isLiteral]  # type: List[GlobBlock]  # source literals  # line 408
        nextliteral = 0  # type: int  # line 409
        parsedOld = []  # type: List[GlobBlock2]  # line 410
        index = 0  # type: int  # line 411
        for part in oldPattern:  # match everything in the old filename  # line 412
            if part.isLiteral:  # line 413
                parsedOld.append(GlobBlock2(True, part.content, part.content))  # line 413
                index += len(part.content)  # line 413
                nextliteral += 1  # line 413
            elif part.content.startswith("?"):  # line 414
                parsedOld.append(GlobBlock2(False, part.content, filename[index:index + len(part.content)]))  # line 414
                index += len(part.content)  # line 414
            elif part.content.startswith("["):  # line 415
                parsedOld.append(GlobBlock2(False, part.content, filename[index]))  # line 415
                index += 1  # line 415
            elif part.content == "*":  # line 416
                if nextliteral >= len(literals):  # line 417
                    parsedOld.append(GlobBlock2(False, part.content, filename[index:]))  # line 417
                    break  # line 417
                nxt = filename.index(literals[nextliteral].content, index)  # type: int  # also matches empty string  # line 418
                parsedOld.append(GlobBlock2(False, part.content, filename[index:nxt]))  # line 419
                index = nxt  # line 419
            else:  # line 420
                Exit("Invalid file pattern specified for move/rename")  # line 420
        globs = [g for g in parsedOld if not g.isLiteral]  # type: List[GlobBlock2]  # line 421
        literals = [l for l in newPattern if l.isLiteral]  # target literals  # line 422
        nextliteral = 0  # line 423
        nextglob = 0  # type: int  # line 423
        outname = []  # type: List[str]  # line 424
        for part in newPattern:  # generate new filename  # line 425
            if part.isLiteral:  # line 426
                outname.append(literals[nextliteral].content)  # line 426
                nextliteral += 1  # line 426
            else:  # line 427
                outname.append(globs[nextglob].matches)  # line 427
                nextglob += 1  # line 427
        pairs.append((filename, "".join(outname)))  # line 428
    return pairs  # line 429

@_coconut_tco  # line 431
def reorderRenameActions(actions: '_coconut.typing.Sequence[Tuple[str, str]]', exitOnConflict: 'bool'=True) -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 431
    ''' Attempt to put all rename actions into an order that avoids target == source names.
      Note, that it's currently not really possible to specify patterns that make this work (swapping "*" elements with a reference).
      An alternative would be to always have one (or all) files renamed to a temporary name before renaming to target filename.
  '''  # line 435
    if not actions:  # line 436
        return []  # line 436
    sources = None  # type: List[str]  # line 437
    targets = None  # type: List[str]  # line 437
    sources, targets = [list(l) for l in zip(*actions)]  # line 438
    last = len(actions)  # type: int  # line 439
    while last > 1:  # line 440
        clean = True  # type: bool  # line 441
        for i in range(1, last):  # line 442
            try:  # line 443
                index = targets[:i].index(sources[i])  # type: int  # line 444
                sources.insert(index, sources.pop(i))  # bubble up the action right before conflict  # line 445
                targets.insert(index, targets.pop(i))  # line 446
                clean = False  # line 447
            except:  # target not found in sources: good!  # line 448
                continue  # target not found in sources: good!  # line 448
        if clean:  # line 449
            break  # line 449
        last -= 1  # we know that the last entry in the list has the least conflicts, so we can disregard it in the next iteration  # line 450
    if exitOnConflict:  # line 451
        for i in range(1, len(actions)):  # line 452
            if sources[i] in targets[:i]:  # line 453
                Exit("There is no order of renaming actions that avoids copying over not-yet renamed files: '%s' is contained in matching source filenames" % (targets[i]))  # line 453
    return _coconut_tail_call(list, zip(sources, targets))  # line 454

def relativize(root: 'str', path: 'str') -> 'Tuple[str, str]':  # line 456
    ''' Determine OS-independent relative folder path, and relative pattern path. '''  # line 457
    relpath = os.path.relpath(os.path.dirname(os.path.abspath(path)), root).replace(os.sep, SLASH)  # line 458
    return relpath, os.path.join(relpath, os.path.basename(path)).replace(os.sep, SLASH)  # line 459

def parseOnlyOptions(root: 'str', options: '_coconut.typing.Sequence[str]') -> 'Tuple[_coconut.typing.Optional[FrozenSet[str]], _coconut.typing.Optional[FrozenSet[str]]]':  # line 461
    ''' Returns set of --only arguments, and set or --except arguments. '''  # line 462
    cwd = os.getcwd()  # type: str  # line 463
    onlys = []  # type: List[str]  # line 464
    excps = []  # type: List[str]  # line 464
    index = 0  # type: int  # line 464
    while True:  # line 465
        try:  # line 466
            index = 1 + listindex(options, "--only", index)  # line 467
            onlys.append(options[index])  # line 468
        except:  # line 469
            break  # line 469
    index = 0  # line 470
    while True:  # line 471
        try:  # line 472
            index = 1 + listindex(options, "--except", index)  # line 473
            excps.append(options[index])  # line 474
        except:  # line 475
            break  # line 475
    return (frozenset((relativize(root, o)[1] for o in onlys)) if onlys else None, frozenset((relativize(root, e)[1] for e in excps)) if excps else None)  # line 476
